#!/usr/bin/perl
# 'Simple' IRC bot to proxy message from this machine to server/channel(s) specified below.
# TODO:
# - stats for past day (problems seen through input, acks, etc)
use strict;
use warnings;
use IO::Socket::UNIX qw/SOCK_STREAM/;

$|=1;
print "$$ ***************** Starting bot at ".localtime(time)." *****************\n";
use POE;
use POE::Component::IRC::State;
use POE::Component::IRC::Plugin::AutoJoin;
use POE::Component::IRC::Plugin::FollowTail;
use POE::Component::IRC::Plugin::Connector;
use POSIX qw(mkfifo);
use Data::Dumper;
use Config::Tiny;

my $cfgfile = 'ircbot.cfg';
my $config = Config::Tiny->read( $cfgfile, 'utf8' );
unless(defined($config))
{
	die "Could not load config from $cfgfile: ".Config::Tiny->errstr();
}

# Get config variables
my $readpipe = $config->{paths}{msgpipe} // die "Config incomplete! Configuration [paths]->'msgpipe' not defined!";
my $rwLive = $config->{paths}{livesocket} // die "Config incomplete! Configuration [paths]->'livesocket' not defined!";
my $lag_o_meter = $config->{log}{lagmeter} // 1;
my $tailRegistered = 0;

my @connections;
foreach my $i ("", 1..9)
{
	my $key = "irc".$i;
	if (defined($config->{$key}))
	{
		my $nickname = $config->{$key}{nickname} // die "Config incomplete! Configuration [$key]->'nickname' not defined!";
		my $ircname  = $config->{$key}{ircname} // die "Config incomplete! Configuration [$key]->'ircname' not defined!";
		my $server = $config->{$key}{server} // die "Config incomplete! Configuration [$key]->'server' not defined!";
		my $ssl = $config->{$key}{ssl} // die "Config incomplete! Configuration [$key]->'ssl' not defined!";
		my $port = $config->{$key}{port} // die "Config incomplete! Configuration [$key]->'port' not defined!";
		my $ipv6 = $config->{$key}{ipv6} // 1;
		my $color = $config->{$key}{colors} // 1;
		my $debug = $config->{$key}{debug} // 0;
		my %connect = (
			nick => $nickname,
			ircname => $ircname,
			server => $server,
			UseSSL => $ssl,
			port => $port,
			useipv6 => $ipv6,
			debug => $debug,
		);
		if (defined($config->{$key}{'username'}) && length($config->{$key}{'username'}))
		{
			$connect{Username} = $config->{$key}{username};
		}
		if (defined($config->{$key}{'password'}) && length($config->{$key}{'password'}))
		{
			$connect{Password} = $config->{$key}{password};
		}

		my $regonly = $config->{$key}{regonly};
		my %channels = ();
		$key = "channels".$i;
		foreach my $chan (keys %{$config->{$key}})
		{
			$channels{'#'.$chan} = $config->{$key}{$chan};
		}
		my %regchannels = %channels;
		push @connections, { connect => \%connect, channels => \%channels, regchannels => \%regchannels, regonly => ($regonly // 0), colors => $color};
	}
}
# die Dumper \@connections;
die "Config incomplete! No irc blocks defined! Make sure you have at least one irc[number] section, such as 'irc1' with nickname, ircname, server, ssl and port." unless(scalar(@connections));

my %c = (
	'/'      => "\e[0m",
	'black'  => "\e[30m",   'on_black'   => "\e[40m",
	'r'      => "\e[31m",   'on_red'     => "\e[41m",
	'R'      => "\e[31;1m",
	'g'      => "\e[32m",   'on_green'   => "\e[42m",
	'G'      => "\e[32;1m",
	'y'      => "\e[33m",   'on_yellow'  => "\e[43m",
	'Y'      => "\e[33;1m",
	'b'      => "\e[34m",   'on_blue'    => "\e[44m",
	'B'      => "\e[34;1m",
	'm'      => "\e[35m",
	'M'      => "\e[35;1m",
	'c'      => "\e[36m",   'on_cyan'    => "\e[46m",
	'C'      => "\e[36;1m",
	'w'      => "\e[37m",   'on_white'   => "\e[47m",
	'W'      => "\e[37;1m"
);

my %colorCodes = (
	'0'   => "\017",        # default colors
	'31'   => "\00305", # brown
	'32'   => "\00303", # green
	'33'   => "\00307", # orange
	'34'   => "\00302", # blue
	'35'   => "\00306", # purple
	'36'   => "\00310", # cyan
	'37'   => "\00315", # white
	'4'   => "\037",        # underline
	bold => {
		'0'   => "\017",        # uuhh
		'31'   => "\00304", # red
		'32'   => "\00309", # green
		'33'   => "\00308", # yellow
		'34'   => "\00312", # blue
		'35'   => "\00313", # purple
		'36'   => "\00311", # cyan
		'37'   => "\00300", # white
		'4'   => "\037",
	}
);

sub translateColors
{
	my ($s) = @_;
	my $boldEnabled = 0;    # Note, maybe change this so there is no bold 'state', but just pick a color.
	while ($s =~ m#(\e\[(([0-9]+|;)+)m)#)
	{
		my $matched = $1;
		my $c = $2; # colors etc
		my @tmp = split /;/, $c;
		my %active;
		foreach (@tmp)
		{
			if ($_ == 1)
			{
				$boldEnabled = 1;
				next;
			}
			if ($_ == 0)
			{
				$boldEnabled = 0;
			}
			if (defined($colorCodes{$_}))
			{
				$active{$_} = 1;
			}
			else
			{
				debug("Color code $_ not translatable (yet) -- please add.");
			}
		}
		my $ccRef = ($boldEnabled ? $colorCodes{bold} : \%colorCodes);
		my $str = join("", map { $ccRef->{$_} } keys %active) // "";
		$s =~ s/\Q$matched\E/$str/;     # quoted, so should always kill the while loop on the found part (and no weird stuff).
	}
	$s =~ s#\003[0-9]{2},\K#01,#g;  # Add ,00 to where another ,number might follow...
	return $s;
}

# Check if the pipe exists
if (-e $readpipe)
{
	# Make sure this is a fifo, if it's a regular file we'll end up with buffering issues
	unlink($readpipe) or die "Could not remove existing pipe $readpipe: $!";
}
unless (-p $readpipe)
{
	mkfifo($readpipe, 0770) or die "Could not create pipe for reading at $readpipe: $!";
}

my %sessions;
my @ircs;
foreach my $con (@connections)
{
	my $irc = POE::Component::IRC::State->spawn( %{$con->{"connect"}}) or die "Could not connect to ".$con->{"connect"}{"server"}."! $!";
	print "Created IRC object for server ".$con->{"connect"}->{"server"}."\n";
	POE::Session->create(
		package_states => [
			main => [ qw(_default _start irc_001 irc_join irc_tail_input irc_tail_error irc_tail_reset irc_msg irc_public input_event lag_o_meter) ],
		],
		heap => { irc => $irc, channels => $con->{"channels"}, connect => $con->{"connect"}, regchannels => $con->{"regchannels"}, regonly => $con->{"regonly"} },
	);
	print "Object created\n";
	if ($con->{"regonly"})
	{
		print "This IRC connection is active on registered channels due to config 'regonly' setting enabled. These are: ".join(", ", keys %{$con->{"regchannels"}}).$/;
	}
}

$poe_kernel->run();

sub _start
{
	my ($heap, $kernel) = @_[HEAP, KERNEL];
	print "Global _start for session ".$_[SESSION]->ID().$/;
	$sessions{$_[SESSION]->ID()} = $_[SESSION]->ID();
	my $irc = $heap->{irc};
	my $channels = $heap->{channels};
	my $connect = $heap->{connect};
	print "Adding channels to join: ".join(", ", sort keys %$channels).$/;
	$irc->plugin_add('AutoJoin', POE::Component::IRC::Plugin::AutoJoin->new( Channels => $channels, RejoinOnKick => 1, Rejoin_delay => 5, Retry_when_banned => 10 ));
	$heap->{connector} = POE::Component::IRC::Plugin::Connector->new(reconnect => 5, delay => 30); # wait 5 secs before reconnect, 30 secs between pings
	$irc->plugin_add( 'Connector' => $heap->{connector} );
	# $irc->yield(register => qw(join irc_tail_input));
	$irc->yield ( connect => $connect );
	$kernel->delay( 'lag_o_meter' => 60 ) if ($lag_o_meter);
	$irc->yield(register => qw(all));
	$irc->yield(connect => { } );
	print "Registering signal handlers for session ".$_[SESSION]->ID().$/;
	$kernel->sig( tailinput => "input_event" );
	$kernel->sig( INT => "int_handler" );
	unless ($tailRegistered)
	{
		print "Registering readpipe at $readpipe\n";
		$irc->plugin_add( 'FollowTail' => POE::Component::IRC::Plugin::FollowTail->new(
				filename => $readpipe,
			));
		$tailRegistered = 1;
	}
	return;
}

sub int_handler
{
	my ($kernel, $heap) = @_[KERNEL, HEAP];
	print "Int handler!\n";
	$heap->{irc}->yield('quit', 'Bye!');
	$heap->{shutting_down} = 1;
	$kernel->sig_handled();
}

sub irc_public
{
	my ($heap, $kernel, $sender, $who, $where, $msg) = @_[HEAP, KERNEL, SENDER, ARG0, ARG1, ARG2];
	my $nick    = (split /!/, $who)[0];
	my $channel = $where->[0];
	my $irc = $sender->get_heap();
	my $muted = $heap->{'muted'};
	my $server = $irc->server_name();
	print "<$nick:$channel@".$server."> $msg\n";

	if ($heap->{"regonly"})
	{
		# See if we care about this channel.
		return unless (grep { $channel eq $_ } keys %{$heap->{"regchannels"}}); # Nope filter :)
	}

	my $nickname = $irc->nick_name();
	if ($msg =~ /^$nickname\s*:\s*(.+)\s*$/i)
	{
		my $cmd = $1;
		if ($cmd =~ /^ack\s+all(\s+(?<what>[^\|]+?)\s*)?\s*(\|\|\s*(?<comment>.+))?$/)
		{ # ack all uil.winnipeg.nl || alles is stuk
			my $sticky = 1;
			my $notify = 1;
			my $persist = 0;
			my $author = $nick;
			my $comment = $+{'comment'} // 'Ack';
			my $what = $+{'what'};
			$comment =~ s/^\s+|\s+$//g;

			print " -> Received command: ack all author[$author] what[$what] comment[$comment]\n";
			# Fetch unacked services, filter on $what
			my $problems = getProblemServices($what);
			my $i = 0;
			# cols = qw(host_name service_description state state_type last_state_change notes comments);
			foreach my $p (@$problems)
			{
				my $cmd = 'ACKNOWLEDGE_SVC_PROBLEM';
				my $r = cmkLiveCommand($cmd, $p->{'host_name'}, $p->{'service_description'}, $sticky, $notify, $persist, $author, $comment);
				if ($r == 1)
				{
					$irc->yield(privmsg => $channel, translateColors($c{'r'}."Error".$c{'/'}." sending command to Check_MK! ".$c{'r'}."NOT ACKNOWLEDGED!".$c{'/'}));
				}
				else
				{
					my $resmsg = "Sent acknowledge command for [".$c{'y'}.$p->{'host_name'}.$c{'/'}."] ".$c{'b'}.$p->{"service_description"}.$c{'/'}." (".$c{'y'}."$comment".$c{'/'}.")";
					$irc->yield(privmsg => $channel, translateColors($resmsg));
				}
			}
		}
		elsif ($cmd =~ /^(?<cmd>[^\s]+)\s+\[(?<host>[^\]]+)\]\s+(?<service>[^\|]+)\s*(\|\|\s*(?<comment>.+)\s*)?$/)
		{ # NagiosBot: ack [uil.winnipeg.nl] mtr to www.google.com || Test
			# my ($cmd, $host, $service, $sticky, $notify, $persist, $author, $comment) = @_;
			my $cmd = $+{'cmd'};
			my $host = $+{'host'};
			my $service = $+{'service'};
			my $sticky = 1;
			my $notify = 1;
			my $persist = 0;
			my $author = $nick;
			my $comment = $+{'comment'} // 'Ack';
			$service =~ s/^\s+|\s+$//g; # Get rid of trailing/leading whitespace
			$comment =~ s/^\s+|\s+$//g;
			print " -> Received command: $cmd host[$host] author[$author] service[$service] comment[$comment]\n";
			if ($cmd =~ /ack/i)
			{
				$cmd = 'ACKNOWLEDGE_SVC_PROBLEM';
			}
			my $r = cmkLiveCommand($cmd, $host, $service, $sticky, $notify, $persist, $author, $comment);
			if ($r == 1)
			{
				$irc->yield(privmsg => $channel, translateColors($c{'r'}."Error".$c{'/'}." sending command to Check_MK! ".$c{'r'}."NOT ACKNOWLEDGED!".$c{'/'}));
			}
			else
			{
				my $resmsg = "Sent acknowledge command for [".$c{'y'}.$host.$c{'/'}."] ".$c{'b'}.$service.$c{'/'}." (".$c{'y'}."$comment".$c{'/'}.")";
				$irc->yield(privmsg => $channel, translateColors($resmsg));
			}
		}
		elsif ($cmd =~ /^problems?\s*(?<filter>[^\s]+)?\s*$/)
		{
			my $filter = $+{'filter'};
			$filter =~ s/^\s+|\s+$//g if(defined($filter));
			$filter = undef unless(defined($filter) && length($filter));
			print " -> request for open problems filtered on [$filter]\n";
			my $problems = getProblemServices($filter);
			my $header = "Unacknowledged services ".($filter?"with filter ".$c{'y'}.$filter.$c{'/'}." ":"")."ordered by last state change";
			if (!defined($filter) && scalar(@$problems) > 8)
			{
				my $orgcount = scalar(@$problems);
				splice(@$problems, 0, 8);
				$header .= " (showing only first 8 results out of $orgcount!)";
			}
			$irc->yield(privmsg => $channel, translateColors($header.":"));
			foreach my $p (@$problems)
			{
				my $sc = getStateColor($p->{"state"});
				my $line = sprintf("[".$c{'y'}."%s".$c{'/'}."] $c{'b'}\%s$c{'/'} %s%s$c{'/'} since $c{'y'}\%s$c{'/'}", $p->{"host_name"}, $p->{"service_description"}, $sc, $p->{"state"}, scalar(localtime($p->{"last_state_change"})));
				$irc->yield(privmsg => $channel, translateColors($line));
			}
		}
		elsif ($cmd =~ /^unmute\s*$/)
		{
			my $msg = "MUTE ".$c{'r'}."DISABLED".$c{'/'}." - messages will be relayed again immediately!";
			$irc->yield(privmsg => $channel, translateColors($msg));
			$heap->{muted} = 0;
		}
		elsif ($cmd =~ /^mute?\s*(?<time>\d*)?\s*$/)
		{
			my $muteDuration = ($+{'time'} // 300);
			$heap->{muted} = time + $muteDuration;
			my $msg = "MUTE ".$c{'y'}."ENABLED".$c{'/'}." - will not relay messages for ".$c{'y'}.$muteDuration.$c{'/'}." seconds!";
			$irc->yield(privmsg => $channel, translateColors($msg));
		}
		elsif ($cmd =~ /^(help|\?)$/)
		{
			$irc->yield(privmsg => $channel, translateColors("You requested help! Available help:"));
			help($irc, $channel);
		}
		else
		{
			$irc->yield(privmsg => $channel, translateColors("Invalid command or syntax error!"));
			help($irc, $channel);
		}
	}
}

sub help
{
	my ($irc, $channel) = @_;
	$irc->yield(privmsg => $channel, translateColors("Acknowledge service problems with: ack [hostname] service name || Comment."));
	$irc->yield(privmsg => $channel, translateColors("Other commands: 'mute [seconds]', 'unmute', 'problems [filter]' and 'ack all [filter]'"));
	$irc->yield(privmsg => $channel, translateColors(" * Filter can be either ".$c{'y'}."a string".$c{'/'}." OR ".$c{'y'}."key=regex;key2=regex".$c{'/'}." where key can be a (partial) match for "));
	$irc->yield(privmsg => $channel, translateColors($c{'y'}.join($c{'/'}.", ".$c{'y'}, qw(host_name service_description state state_type last_state_change notes comments contact_groups)).$c{'/'}));
	$irc->yield(privmsg => $channel, translateColors("Example: ".$irc->nick_name().": ".$c{'g'}."ack all host=foo;service=ntp".$c{'/'}." will ack all services that have ntp in their description on host(s) matching foo"));
}

sub getStateColor
{
	my ($state) = @_;
	if ($state =~ /ok/i)
	{
		return $c{'g'};
	}
	elsif ($state =~ /warn/i)
	{
		return $c{'y'};
	}
	elsif ($state =~ /crit/i)
	{
		return $c{'r'};
	}
	return $c{'/'};
}

sub irc_msg {
	# ARG0 is the nick!hostmask of the sender.
	# ARG1 is an array reference containing the nick(s) of the recipients.
	# ARG2 is the text of the message. On servers supporting the IDENTIFY-MSG feature (e.g. FreeNode), there will be an additional argument,
	# ARG3, which will be 1 if the sender has identified with NickServ, 0 otherwise.

	my ($sender, $msg) = @_[SENDER, ARG2];
	my $irc = $sender->get_heap();
	print "Message from $sender: $msg\n";
	return;
}

sub irc_001 {
	my $sender = $_[SENDER];

	# Since this is an irc_* event, we can get the component's object by
	# accessing the heap of the sender. Then we register and connect to the
	# specified server.
	my $irc = $sender->get_heap();

	print "Connected to ", $irc->server_name(), "\n";

	# we join our channels
	# $irc->yield( join => $_ ) for keys %channels;
	return;
}

sub irc_join {
	my $nick = (split /!/, $_[ARG0])[0];
	my $channel = $_[ARG1];
	my $irc = $_[SENDER]->get_heap();

	# Figure out what IRC server this is
	my $heap = $_[HEAP];
	#print Dumper $heap->{regchannels};
	#print "Server: ".$irc->{server}. " / ".$irc->server_name().$/;
	#print "Complete heap dump:\n";
	#print Dumper $heap;

	# only send the message if we were the one joining
	if ($nick eq $irc->nick_name()) {
		if ($heap->{'regonly'})
		{
			return unless (grep { $channel eq $_ } keys %{$heap->{"regchannels"}}); # Don't talk if not allowed.
		}
		$irc->yield(privmsg => $channel, $nick.' reporting for duty!');
	}
}

sub input_event
{
	my ($session, $heap, $kernel, $sender, $event, $callsession, $filename, $input) = @_[SESSION, HEAP, KERNEL, SENDER, ARG0, ARG1, ARG2, ARG3];
	# print "Global input_event\n";
	my $channels = $heap->{channels}; # don't broadcast to invited etc channels
	$channels = $heap->{regchannels} if ($heap->{regonly});
	#print "Channels\n";
	#print Dumper $channels;
	my $irc = $heap->{irc};
	if (time > ($heap->{muted} // 0))
	{
		print "Sending input to IRC channels ".join(", ", keys %$channels)." on server ".$irc->server_name()." from [$filename]: $input\n";
		# $kernel->post( $sender, 'privmsg', $_, $input ) for keys %$channels;
		foreach (keys %$channels)
		{
			$irc->yield(privmsg => $_, translateColors($input));
		}
	}
	else
	{
		print "NOT Sending input to IRC due to mute on channels ".join(", ", keys %$channels)." on server ".$irc->server_name()." from [$filename]: $input\n";
	}
	return;
}

sub irc_tail_input {
	my ($heap, $kernel, $sender, $filename, $input) = @_[HEAP, KERNEL, SENDER, ARG0, ARG1];
	print "Tail Input [$filename]: $input -- BROADCASTING\n";
	broadcast('tailinput', $sender, $filename, $input);
	return;
}

sub irc_tail_error {
	my ($kernel, $sender, $filename, $errnum, $errstring) = @_[KERNEL, SENDER, ARG0 .. ARG2];
	# print "$filename: ERROR: $errnum $errstring\n";
	#$kernel->post( $sender, 'privmsg', $_, "$filename: ERROR: $errnum $errstring" ) for @channels;
	# $irc->plugin_del( 'FollowTail' );
	return;
}

sub irc_tail_reset {
	my ($kernel, $sender, $filename) = @_[KERNEL, SENDER, ARG0];
	print "Reset [$filename]\n";
	# $kernel->post( $sender, 'privmsg', $_, "$filename: RESET EVENT" ) for keys %channels;
	return;
}

sub _default {
	my ($event, $args) = @_[ARG0 .. $#_];
	my @output = ( "$event: " );

	for my $arg (@$args) {
		if ( ref $arg eq 'ARRAY' ) {
			push( @output, '[' . join(', ', @$arg ) . ']' );
		}
		else {
			push ( @output, "'".($arg//'undef')."'" );
		}
	}
	# print "_default: ".join(", ", @_).$/;
	# print join ' ', @output, "\n";
	return;
}

sub cmkLiveCommand
{ # https://old.nagios.org/developerinfo/externalcommands/commandinfo.php?command_id=40
	my ($cmd, $host, $service, $sticky, $notify, $persist, $author, $comment) = @_;
	# ACKNOWLEDGE_SVC_PROBLEM;<host_name>;<service_description>;<sticky>;<notify>;<persistent>;<author>;<comment>
	# ACKNOWLEDGE_SVC_PROBLEM;host1;service1;2;1;1;Some One;Some Acknowledgement Comment\n
	unless (-S $rwLive)
	{
		print "Error: [$rwLive] is not a socket!\n";
		return 1;
	}

	my $socket = IO::Socket::UNIX->new(
		Type => SOCK_STREAM,
		Peer => $rwLive,
	);
	unless(defined($socket))
	{
		print "Error opening Live socket [$rwLive]: $!\n";
		return 1;
	}
	my $command = sprintf("COMMAND [%s] %s;%s;%s;%s;%s;%s;%s;%s\n\n", time, $cmd, $host, $service, $sticky, $notify, $persist, $author, $comment);
	print "Sending command to rwLive socket: $command";
	print $socket $command;

	close($socket);
	return 0;
}

sub getProblemServices
{
	my ($filter) = $1;
	$filter =~ s/^\s+//g;
	$filter =~ s/\s+$//g;
	print "getProblemServices($filter) called\n";
	unless (-S $rwLive)
	{
		print "Error: [$rwLive] is not a socket!\n";
		return 1;
	}
	my $socket = IO::Socket::UNIX->new(
		Type => SOCK_STREAM,
		Peer => $rwLive,
	);
	unless(defined($socket))
	{
		print "Error opening Live socket [$rwLive]: $!\n";
		return 1;
	}

	my @cols = qw(host_name service_description state state_type last_state_change notes comments contact_groups);
	my $command = "GET services
Columns: ".join(" ", @cols)."
Filter: state = 1
Filter: acknowledged = 0
And: 2
Filter: state = 2
Filter: acknowledged = 0
And: 2
Or: 2
\n";
	print "Sending command to rwLive socket: $command";
	print $socket $command;

	# See for result
	my %states = (
		0 => 'OK',
		1 => 'WARNING',
		2 => 'CRITICAL',
		3 => 'UNKNOWN'
	);
	my $res = [];
	# Assume filter is either a word like a hostname or servicename, OR a targeted something in the syntax of: "host=uil;contact_groups=SLA" - note that we'll regex against the columns, so cont=bla will work but match both contacts and contact_groups. Since we only request the contact_groups col it doesn't matter though ;)
	# In the ; syntax we'll use AND for multiple supplied cols.
	my %fgroups;
	if (defined($filter))
	{
		my %cgroups = map { split m{\s*=\s*} } split m{\s*;\s*}, $filter;
		# See if host/service matches this
		foreach my $k (keys %cgroups)
		{
			# Matches specific key
			foreach my $col (grep /$k/i, @cols)
			{
				$fgroups{$col} = $cgroups{$k};
			}
		}
	}
	while(<$socket>)
	{
		my $entry = {};
		my @names = @cols;
		foreach my $e (split/;/)
		{
			my $n = shift @names;
			if ($n eq 'state')
			{
				$e = $states{$e} if(defined($states{$e}));
			}
			$entry->{$n} = $e;
		}
		if (defined($filter) || scalar(keys(%fgroups)))
		{
			# See if host/service matches this
			if (scalar(keys %fgroups))
			{
				my $match = 1;
				foreach my $k (keys %fgroups)
				{
					unless ($entry->{$k} =~ /$fgroups{$k}/i)
					{
						$match = 0;
					}
				}
				push @$res, $entry if ($match == 1);
			}
			elsif ($entry->{"service_description"} =~ /$filter/i || $entry->{"host_name"} =~ /$filter/i || $entry->{"notes"} =~ /$filter/i ||$entry->{"comments"} =~ /$filter/i)
			{
				push @$res, $entry;
			}
			else
			{
				# print " - skipping entry $entry->{"host_name"} / $entry->{"service_description"}, no filter match.\n";
			}
		}
		else
		{
			push @$res, $entry;
		}
	}
	@$res = sort { $a->{'last_state_change'} <=> $b->{'last_state_change'} } @$res;
	# print "getProblemServices result set:\n"; print Dumper $res;
	return $res;
}

sub lag_o_meter {
	my ($kernel,$heap) = @_[KERNEL,HEAP];
	my $irc = $heap->{irc};
	print 'Time: ' . time() . ' Nick: '.$irc->nick_name().' @ Server: ' . $irc->server_name() . ' Lag: ' . $heap->{connector}->lag() . "\n";
	$kernel->delay( 'lag_o_meter' => 300 );
	return;
}

# Send an event to every registered session.
sub broadcast
{
	my ($event, @etc) = @_;
	foreach (keys %sessions)
	{
		#print "Posting event $event to $_ with args ".join(", ", @etc).$/;
		$poe_kernel->signal($_, $event, @etc) or print "Error signaling to session $_: $!";
	}
}
